#define comment(x) asm ("; " ## x)

#define update_ZSP(y, T) \
  cc_ZF = (y == 0); \
  cc_SF = (y < 0); \
  cc_PF = parity [((unsigned T)y >> 24) ^ ((y >> 16) & 0xff) ^ ((y >> 8) & 0xff) ^ (y & 0xff)]

#define adcl (x, y) \
  tmp = y; \
  y += x + cc_CF; \
  update_ZSP (y, long)

#define add(x, y, T) \
  comment ("add." ## #T ## " " ## #x ## "," ## #y); \
  tmp = y; \
  y += x; \
  update_ZSP (y, T); \
  cc_OF = (x < 0 == tmp < 0) && (y < 0 == x < 0); \
  cc_CF = ((unsigned T)y < (unsigned T)x)

#if 0
#define add(x, y, T) \
  tmp = y; \
  y += x; \
  cc_OF = (tmp < 0 == x < 0) && (x < 0 == y < 0); \
  update_ZSP (y, T); \
  cc_CF = ((unsigned T)y < (unsigned T)x)
#endif

#define addl(x, y) add (x, y, long)

#define andl(x, y) \
  tmp = y; \
  y &= x; \
  update_ZSP (y, long)

#define bswap(x) \
  x = (unsigned long)x >> 24 | (x >> 16) & 0xff | (x >> 8) & 0xff | x & 0xff

#define call(x) \
  comment ("call " ## #x); \
  x ()

#define clc \
  cc_CF = 0

#define cld \
  cc_DF = 0

#define cmc \
  cc_CF = !cc_CF

#define cmp(x,y) \
  subl(x,y); \
  y = tmp

#define dec(x) \
  subl (1, x); \
  cc_ZF = (y == 0); \
  cc_SF = ((signed long)y < 0); \
  cc_OF = (x < 0 == tmp < 0) && (x < 0 != y < 0); \
  cc_PF = parity [((unsigned long)y >> 24) ^ ((y >> 16) & 0xff) ^ ((y >> 8) & 0xff) ^ (y & 0xff)]

#define div(x)

#define enter(x) \
  pushl (ebp); \
  ebp = esp - x

#define inc(x) \
  addl (1, x); \
  cc_ZF = (y == 0); \
  cc_SF = ((signed long)y < 0); \
  cc_OF = (x < 0 == tmp < 0) && (x < 0 != y < 0); \
  cc_PF = parity [((unsigned long)y >> 24) ^ ((y >> 16) & 0xff) ^ ((y >> 8) & 0xff) ^ (y & 0xff)]

#define jxx(x, y) \
  comment ("jxx(" ## #x ## ") " ## #y); \
  if (x) goto y

#define ja(x) jxx (!cc_CF && !cc_ZF, x)
#define jae(x) jnc (x)
#define jb(x) jc (x)
#define jbe(x) jxx (cc_CF || cc_ZF, x)
#define jc(x) jxx (cc_CF, x)
#define jcxz(x) jxx (cx == 0, x)
#define jecxz(x) jxx (ecx == 0, x)
#define je(x) jz (x)
#define jg(x) jxx (!cc_ZF && (cc_SF == cc_OF), x)
#define jge(x) jxx (cc_SF == cc_OF, x)
#define jl(x) jxx (cc_SF != cc_OF, x)
#define jle(x) jxx (cc_ZF || (cc_SF != cc_OF), x)
#define jmp(x) jxx (1, x)
#define jna(x) jxx (cc_CF || cc_ZF, x)
#define jnae(x) jc (x)
#define jnb(x) jnc (x)
#define jnbe(x) jxx (!cc_CF && !cc_ZF, x)
#define jnc(x) jxx (!cc_CF, x)
#define jne(x) jnz (x)
#define jng(x) jxx (cc_ZF || (cc_SF != cc_OF), x)
#define jnge(x) jxx (cc_SF != cc_OF, x)
#define jnl(x) jxx (cc_SF == cc_OF, x)
#define jnle(x) jxx (!cc_ZF && (cc_SF == cc_OF), x)
#define jno(x) jxx (!cc_OF, x)
#define jnp(x) jxx (!cc_P, x)
#define jns(x) jxx (!cc_SF, x)
#define jnz(x) jxx (!cc_ZF, x)
#define jo(x) jxx (cc_OF, x)
#define jp(x) jxx (cc_P, x)
#define jpe(x) jp (x)
#define jpo(x) jnp (x)
#define js(x) jxx (cc_SF, x)
#define jz(x) jxx (cc_ZF, x)

#define lahf \
  ah = cc_CF | cc_PF << 2 | cc_AF << 4 | cc_ZF << 6 | cc_SF << 7

#define leave \
  esp = ebp; \
  popl (ebp)

#define loop(x) \
  cx--; \
  if (cx != 0) \
    goto x

#define loopz(x) \
  cx--; \
  if (cx != 0 && cc_ZF) \
    goto x

#define loope(x) loopz (x)

#define loopnz(x) \
  cx--; \
  if (cx != 0 && !cc_ZF) \
    goto x

#define loopne(x) loopnz (x)

#define movl(x,y) \
  y = x

#define mulb(x) \
  ax = x * al; \
  cc_CF = ((unsigned long)ax < (unsigned long)x); \
  cc_OF = 0

#define mulw(x) \
  eax = x * ax; \
  cc_CF = ((unsigned long)eax < (unsigned long)x); \
  cc_OF = 0

#define mull(x) \
  eax *= x; \
  cc_CF = ((unsigned long)y < (unsigned long)x); \
  cc_OF = 0

#define negl(x) \
  x = -x; \
  cc_ZF = (x == 0); \
  cc_SF = ((signed long)x < 0); \
  cc_CF = 0; \
  cc_OF = 0

#define nop

#define notl(x) \
  x = ~x

#define orl(x, y) \
  y |= x; \
  update_ZSP (y, long)

#define pop(x, T) \
  x = *(T *)esp; \
  esp += sizeof (T)

#define popl(x) pop (x, long)

#define push(x, T) \
  comment ("push." ## #T ## " " ## #x); \
  esp -= sizeof (T); \
  *(T *)esp = x

#define pushl(x) push ((long)x, long)

#define rcl(x, y) \
  tmp = x; \
  x = (x << y) | ((unsigned long)x >> (32 - y)); \
  cc_CF = x & 1; \
  cc_OF = (tmp < 0) != (x < 0)

#define rcr(x, y) \
  tmp = x; \
  x = ((unsigned long)x >> y) | (x << (32 - y)); \
  cc_CF = (x < 0); \
  cc_OF = (tmp < 0) != cc_CF

#define rol(x, y) rcl(x, y)

#define ror(x, y) rcr(x, y)

#define sahf \
  cc_CF = (ah & 0x01) != 0; \
  cc_PF = (ah & 0x04) != 0; \
  cc_AF = (ah & 0x10) != 0; \
  cc_ZF = (ah & 0x40) != 0; \
  cc_SF = (ah & 0x80) != 0

#define sal(x,y) \
  tmp = x; \
  x <<= y; \
  cc_ZF = (x == 0); \
  cc_SF = ((signed long)x < 0); \
  cc_CF = 0; \
  cc_OF = (x < 0 != tmp < 0)

#define sar(x, y) \
  x = (signed long)x >> y; \
  cc_ZF = (x == 0); \
  cc_SF = ((signed long)x < 0); \
  cc_CF = 0; \
  cc_OF = 0

#define sbb(x, Y) \
  tmp = y; \
  y -= x + cc_CF; \
  update_ZSP (y, long)

#define setae(x) \
  x = !cc_CF

#define setnb(x) setae (x)

#define setb(x) \
  x = cc_CF

#define setnae(x) setb (x)

#define setbe(x) \
  x = cc_CF || cc_ZF

#define setna(x) setbe (x)

#define setx(x) \
  x = cc_ZF

#define sete(x) setz (x)

#define setnz(x) \
  x = !cc_ZF

#define setne(x) setnz (x)

#define setl(x) \
  x = cc_SF != cc_OF

#define setnge(x) setl(x)

#define setge(x) \
  x = cc_SF == cc_OF

#define setnl(x) setge (x)

#define setle(x) \
  x = cc_ZF || (cc_SF != cc_OF)

#define setng(x) setle (x)

#define setg(x) \
  x = !cc_ZF || (cc_SF == cc_OF)

#define setnle(x) setg (x)

#define sets(x) \
  x = cc_SF

#define setns(x) \
  x = !cc_SF

#define setc(x) \
  x = cc_CF

#define setnc(x) \
  x = !cc_CF

#define seto(x) \
  x = cc_OF

#define setno(x) \
  x = !cc_OF

#define setp(x) \
  x = cc_PF

#define setpe(x) setp (x)

#define setnp(x) \
  x = !cc_PF

#define setpo(x) setnp (x)

#define shl(x, y) sal (x, y)

#define shr(x, y) \
  tmp = x; \
  x = (unsigned long)x >> y; \
  cc_ZF = (x == 0); \
  cc_SF = ((signed long)x < 0); \
  cc_CF = 0; \
  cc_OF = (x < 0 != tmp < 0)

#define stc \
  cc_CF = 1

#define std \
  cc_DF = 1

#define subl(x, y) \
  comment ("sub.long " ## #x ## "," ## #y); \
  tmp = y; \
  y -= x; \
  update_ZSP (y, long); \
  cc_CF = ((unsigned long)y > (unsigned long)tmp); \
  cc_OF = ((long)tmp < 0 != (long)x < 0) && ((long)x < 0 == (long)y < 0)

#define test(x, y) \
  tmp = y; \
  andl (x, y); \
  y = tmp

#define xchg(x, y) \
  tmp = x; \
  x = y; \
  y = tmp

#define xorl(x, y) \
  comment ("xor.long " ## #x ## "," ## #y); \
  y ^= x; \
  update_ZSP (y, long)
